package com.flyfish.framework.logging.service;

import com.flyfish.framework.domain.base.Domain;
import com.flyfish.framework.domain.po.User;
import com.flyfish.framework.logging.config.LoggingTextRegistry;
import com.flyfish.framework.logging.domain.Log;
import com.flyfish.framework.logging.domain.LogType;
import com.flyfish.framework.service.AuthenticationLogger;
import com.flyfish.framework.service.UserFindService;
import com.flyfish.framework.utils.AuthenticationMessages;
import com.flyfish.framework.utils.UserUtils;
import org.springframework.security.core.AuthenticationException;
import org.springframework.security.core.userdetails.UserDetails;
import org.springframework.stereotype.Service;
import org.springframework.util.MultiValueMap;

import javax.annotation.Resource;
import java.util.Date;
import java.util.Optional;

/**
 * 简单的认证日志
 *
 * @author wangyu
 */
@Service
public class SimpleAuthenticationLogger implements AuthenticationLogger {

    @Resource
    private LogService logService;
    @Resource
    private UserFindService userService;
    @Resource
    private LoggingTextRegistry registry;

    /**
     * 记录成功
     *
     * @param user 用户
     */
    @Override
    public void success(UserDetails user) {
        saveSuccess("LOGIN", "登录成功", user);
    }

    /**
     * 记录失败
     *
     * @param form      表单
     * @param exception 错误异常
     */
    @Override
    public void failure(MultiValueMap<String, String> form, AuthenticationException exception) {
        saveError("LOGIN", "登录失败", form.getFirst("username"), AuthenticationMessages.message(exception));
    }

    /**
     * 记录退出
     *
     * @param user 用户
     */
    @Override
    public void logout(UserDetails user) {
        saveSuccess("LOGOUT", "退出成功", user);
    }

    /**
     * 操作成功记录
     *
     * @param business 业务
     * @param message  信息
     * @param user     用户
     */
    private void saveSuccess(String business, String message, UserDetails user) {
        User raw = UserUtils.toUser(user);
        Log log = new Log();
        log.setType(LogType.AUTHENTICATION);
        log.setSignature(business);
        log.setSuccess(true);
        log.setBody("为了安全，本内容隐藏");
        log.setModule("登录模块");
        log.setBusiness(registry.text(business));
        log.setResponse(message);
        log.setOperator(Optional.ofNullable(raw).map(Domain::getName).orElse("未知"));
        log.setPeriod(0L);
        log.setStartTime(new Date());
        // 写入日志
        logService.create(log);
    }

    /**
     * 保存失败日志
     *
     * @param business 业务
     * @param message  消息
     * @param username 用户名
     * @param error    错误
     */
    private void saveError(String business, String message, String username, String error) {
        Log log = new Log();
        log.setType(LogType.AUTHENTICATION);
        log.setSignature(business);
        log.setSuccess(false);
        log.setBody("某人尝试使用用户名'" + username + "'，密码：******，进行登录");
        log.setModule("登录模块");
        log.setBusiness(registry.text(business));
        log.setError(error);
        log.setResponse(message);
        log.setOperator(userService.findByUsername(username).map(Domain::getName).orElse("未知"));
        log.setPeriod(0L);
        log.setStartTime(new Date());
        // 写入日志
        logService.create(log);
    }

}
